#include "stdafx.h"
#include "clf_application.h"
#include "clf_input.h"
#include "clf_util.h"
#include "clf_random.h"
#include "clf_primitives.h"
#include "clf_math.h"
#include "clf_camera.h"
#include "clf_resources.h"
#include "clf_light.h"
#include "clf_font.h"
#include "clf_shader.h"
#include "clf_skybox.h"
#include "clf_effects.h"
#include "clf_quaternion.h"

#include "clf_world.h"
#include "clf_mesh.h"

#include "airplane.h"

extern HWND  hWnd;						// Window Handle

static int m_nFPS = 0;
static int m_nFrames = 0;											// FPS and FPS Counter
static bool g_pause=true;
static float g_strup_timer = 1.2f;

cCamera world_camera;         // view from the window
clf_stage window_viewport;
cCamera map_camera;           // map of the world
clf_stage map_viewport;
cCamera art_horizon;          // artificial horizon instrument
clf_stage ah_viewport;
cCamera inclination;          // inclination instrument
clf_stage incl_viewport;
cCamera instruments;          // instrument pannel
clf_stage insrt_viewport;

// a light source
clf_light g_light(GL_LIGHT0);

// the shader
clf_shader shader;

clf_skybox g_skybox(800);  // size of skybox is small enough for fog effect!!

// class to handle the fog effect
clf_fog g_fog;

// the game world
clf_world g_w;

clf_ground* g_ground;

airplane ship;
clf_vector3D euler_angles;

struct controlState{
	controlState(){
	}
	float pitch;
	float roll;
	float yoke;
	float flaps;
	float thrust;
}ctrlState;
// constant value used to convert joystick axis values
// in the range -1 to 1, required by the airplane simulator code
// instead of -1000 to 1000, returned by the joystick
const float ctrlLimit = 1000.f;

static void update_user_input(float fElapsed)
{
	static float tsx = 0;
	static float tsy = 0;
	static float tst = 0;
	/////////////////////////////////////////////////////////
	DIMOUSESTATE2* ms = getMouseStateImm();
	if (ms) // ms is NULL if mouse is not captured
	{
		if (ms->lX != 0)
		{
			tsx += ms->lX;
			if (tsx < -ctrlLimit) tsx = -ctrlLimit;
			if (tsx > ctrlLimit) tsx = ctrlLimit;
			ctrlState.roll = (float)tsx / ctrlLimit;
		}
		if (ms->lY != 0)
		{
			tsy += ms->lY;
			if (tsy < -ctrlLimit) tsy = -ctrlLimit;
			if (tsy > ctrlLimit) tsy = ctrlLimit;
			ctrlState.pitch = (float)tsy / ctrlLimit;
		}
		if (isKeyPressed(DIK_Z))
		{
			tst -= fElapsed * 400.f;
		}
		if (isKeyPressed(DIK_A))
		{
			tst += fElapsed * 400.f;
		}
		if (tst < 0.f) tst = 0.f;
		if (tst > 1000.f) tst = 1000.f;
		ctrlState.thrust = tst / ctrlLimit;
	}
	else
		abort();
	/////////////////////////////////////////////////////////
	DIJOYSTATE2* js = getJoystickState();
	if (js) // js is NULL if joystick is not captured
	{
		ctrlState.roll = (float)js->lX / ctrlLimit;
		ctrlState.pitch = (float)js->lY / ctrlLimit;
		ctrlState.thrust = (float)(-js->rglSlider[0] + ctrlLimit)/(2*ctrlLimit);
	}
	/////////////////////////////////////////////////////////
}

void request_initial_settings(sInitialSettings& is)
{
	// enable first part to run in window mode
#if 0
	is.isFullScreen = false;
	is.width = 800;
	is.height = 500;
	is.colorDepth = 32;
#else // or enable second part to run in fuul screen mode
	// find your monitor native resolution and apply it
	is.isFullScreen = true;
	is.width = 1680;
	is.height = 1050;
	is.colorDepth = 32;
#endif
	// do not use mouse
	is.directMouse = true;
	// and use joystick
	is.directJoystick = true;
}

void create_world()
{
	g_ground = new clf_ground();
	g_ground->set_extend(5000, -5000, 5000, -5000);
	g_w.add_object(g_ground);

	building *bld;

	int txs[2];
	txs[0] = clf_loadTexture("build1.tga");
	txs[1] = clf_loadTexture("build2.tga");

	float x,z;
	int i=0;
	int j = 0;
	for (x=-150.f; x<151.f; x+=60.f)
	{
		j = 1-j;
		i = j;
		for (z=-150.f; z<151.f; z+=60.f)
		{
			bld = new building(clf_vector3D(x,7.5f,z), clf_vector3D(15,15,15));
			bld->set_texture(txs[i]);
			g_w.add_object(bld);
			i = 1-i;
		}
	}

}

void destroy_world()
{
	std::list<clf_world_object*>::iterator it;
	for (it=g_w.m_objects.begin(); it!=g_w.m_objects.end(); ++it)
		delete (*it);
}


///////////////////////////////////////////////////////////////////////
void restart_game()
{
	g_pause=true;
	ship.InitializeAirplane();
	g_strup_timer = 0.6f;
}

// this function is called when the program starts to initialize the game
void game_init()
{
	init_math_lib();
	get_font_manager().create_font_2d("Bookman2D", "Bookman Old Style", 16);
	get_font_manager().create_font_3d("Bookman3D", "Bookman Old Style", 20, .2f);

	glEnable (GL_DEPTH_TEST);        // enable depth test
	glEnable(GL_COLOR_MATERIAL);     // use color for drawing the object

	// load shader
	shader.load_shaders("phong_vertex_shader2.glsl", "fragment_shader1.glsl");
	
	// Light model parameters:
	// -------------------------------------------
	// create and position a light source
	g_light.set_ambient(0.25f, 0.25f, 0.25f, 1.f);
	g_light.set_diffuse(0.7f, 0.7f, 0.7f, 1.f);
	g_light.set_specular(1.f, 1.f, 1.f, 1.f);
	g_light.set_position(-100.f, 100.f, 100.f, 0.f);

	// -------------------------------------------
	// Material parameters:
	
	GLfloat material_ambient[] = {0.7f, 0.7f, 0.7f, 1.0f};
	GLfloat material_diffuse[] = {0.4f, 0.4f, 0.4f, 1.0f};
	GLfloat material_specular[] = {0.8f, 0.8f, 0.8f, 1.0f};
	GLfloat material_emission[] = {0.1f, 0.1f, 0.1f, 0.0f};
	GLfloat material_shine = 50.0f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material_ambient);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material_diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, material_emission);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, material_shine);

	g_skybox.load_textures("skybox/up.tga", "skybox/dn.tga", "skybox/ft.tga", "skybox/bk.tga", "skybox/lt.tga", "skybox/rt.tga");

	g_fog.density() = 0.01f;
	g_fog.end() = 2500.f;

	create_world();

	restart_game();

	clf_vector3D x;

	art_horizon.vLocation = clf_vector3D(0, 0, 0);
	art_horizon.vUp = clf_vector3D(0, 0, 1);
	art_horizon.vForward = clf_vector3D(1, 0, 0);

	inclination.vLocation = clf_vector3D(-50,0,0);
	inclination.vUp = clf_vector3D(0, 0, 1);
	inclination.vForward = clf_vector3D(1,0,0);

	instruments.vLocation = clf_vector3D(0,0,0);
	instruments.vUp = clf_vector3D(0, 0, 1);
	instruments.vForward = clf_vector3D(1,0,0);
}

// this function is called when the window is resized
void resize_window (int width, int height)			// reshape the window when it's moved or resized
{
	// setup viewport
	int spc = height/3;
	int ht = spc-10;
	int wd = ht;
	int l = width-wd;

	window_viewport.resize(0,spc,width, height-spc);
	window_viewport.set_perspective(30, 0.1f, 10000.f);

	map_viewport.resize(0,0,width/3, spc-10);
	map_viewport.set_perspective(60, 900.f, 1100.f);

	ah_viewport.resize(l, 0, wd, ht);
	ah_viewport.set_perspective(45, .1f, 50.f);

	l -= (wd+10);

	incl_viewport.resize(l, 0, wd, ht);
	incl_viewport.set_perspective(45, 4.f, 550.f);

	l -= (wd+10);
	insrt_viewport.resize(l, 0, wd, ht);
	insrt_viewport.set_perspective(45, .1f, 20.f);
}

// this function is called upon exit to clear allocated resources
void game_terminate()
{
	destroy_world();
	shader.terminate();
}

// the core of the game. This is called repeatedly to update the state of the game
// fElapsed: is the elapsed time since last call. The time is in seconds (usualy a very small number)
void frame_move(float fElapsed)
{
	if (isKeyPressed(DIK_ESCAPE)) 
		DestroyWindow(hWnd);
	if (isKeyPressed(DIK_R))
	{
		//keyDown['R'] = 0;
		restart_game();
		return;
	}
	if (isKeyPressed(DIK_P))
	{
		//keyDown['P'] = 0;
		g_pause = !g_pause;
		g_strup_timer = -1.f;
	}

	if (g_strup_timer > 0.f)
	{
		g_strup_timer -= fElapsed;
		if (g_strup_timer <= 0.f)
			g_pause = false;
	}
	if (!g_pause)
	{
		float lr = 0.f;
		float ud = 0.f;

		update_user_input(fElapsed);
		lr = ctrlState.roll;
		if (lr < 0)
			ship.RollLeft(-lr);
		else if (lr > 0)
			ship.RollRight(lr);
		/////////////////////////////////////////////////////////
		ud = ctrlState.pitch;
		if (ud < 0)
			ship.PitchDown(-ud);
		else if (ud > 0)
			ship.PitchUp(ud);
		/////////////////////////////////////////////////////////
		ship.SetThrust(ctrlState.thrust);
		/////////////////////////////////////////////////////////

		if (isKeyPressed(DIK_HOME))
		{
			lr = 0.f;
			ship.ZeroAilerons();
		}
		if (isKeyPressed(DIK_END))
		{
			ud = 0.f;
			ship.ZeroElevators();
		}

		ship.StepSimulation(fElapsed);
		g_w.step_simulation(fElapsed);
	}

	clf_vector3D z = ship.GetBodyZAxisVector();
	clf_vector3D x = ship.GetBodyXAxisVector();
	world_camera.vForward = clf_vector3D(x.x, x.z, x.y);
	world_camera.vUp = clf_vector3D(z.x, z.z, z.y);
	x = ship.Airplane.vPosition*0.3f; // convert ft to m
	world_camera.vLocation = clf_vector3D(x.x, x.z, x.y);

	map_camera.vLocation = clf_vector3D(x.x, 1000, x.y);
	map_camera.vForward = clf_vector3D(0, -1, 0);
	x = ship.GetBodyXAxisVector();
	map_camera.vUp = clf_vector3D(x.x, 0, x.y);

	euler_angles = make_euler_angles_from_q(ship.Airplane.qOrientation);

	// frame counting mechanism
	// the higher the frame rate the faster the system
	// if frame rate drops below 30 we are in deep trouble
	// we should either optimize the program or buy a new computer
	static dFloat tot = 0;
	tot += fElapsed;
	m_nFrames++;												// Increment Our frame Counter
	if (tot >= 1.f)
	{
		tot = 0;
		m_nFPS = m_nFrames;
		m_nFrames = 0;

		char txt[200];
		sprintf(txt, "GusOnGames, fps:%d",m_nFPS);
		SetWindowText(hWnd, txt);
	}
}

void frame_render()
{
	static GLuint draw_list = -1;
	// clear the screen
	// drawing the skybox means we do not have to waste time clearing the screen
	glClearColor(0.4f, 0.4f, 0.4f, 1.f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);	// clear screen and stencil buffer

	// viewport and perspective
	window_viewport.prepare();

	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	glPushMatrix();

	// position the viewer
	world_camera.apply();

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glEnable(GL_NORMALIZE);
	// position the light
	g_light.enable();

	// fog
	g_fog.fog_apply();
	// render the skybox
	g_skybox.render(world_camera.vLocation.x, world_camera.vLocation.y, world_camera.vLocation.z);

	// shader
	shader.begin();

	// reapply the light
	g_light.apply();

	// draw world objects
	if (draw_list == -1)
	{
		draw_list = glGenLists(1);
		glNewList(draw_list, GL_COMPILE);
		g_w.render();
		glEndList();
	}
	glCallList(draw_list);

	shader.end();
	glPopMatrix();

	if (g_strup_timer > 0.f)
	{
		clf_font* menu_font = get_font_manager().get_font("Bookman3D");
		menu_font->set_color_RGB(0.5f,.75f,.95f);
		glPushMatrix();
			glTranslatef(0, 0, -10);
			menu_font->print_3d(ALIGN_CENTER, "initializing...");
		glPopMatrix();
	}

	glPopAttrib();

	/////////////////////////////////////////////
	// draw the map
	map_viewport.prepare();

	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	glPushMatrix();

	// position the viewer
	map_camera.apply();

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glDisable(GL_DEPTH_TEST);

	glCallList(draw_list);

	glPopMatrix();
	glPopAttrib();

	/////////////////////////////////////////////
	// draw the artificial horizon
	glClearColor(0.f, 0.1f, 0.75f, 1.0f );
	glScissor(ah_viewport.left, ah_viewport.bottom, ah_viewport.width, ah_viewport.height);
	glEnable(GL_SCISSOR_TEST);
	glClear(GL_COLOR_BUFFER_BIT);
	glDisable(GL_SCISSOR_TEST);

	ah_viewport.prepare();
	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	glPushMatrix();

	// position the viewer
	art_horizon.apply();

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glDisable(GL_DEPTH_TEST);

	glPushMatrix();
		glColor3f(.5f, .12f, .0f);
		glRotatef(-euler_angles.y, 0,1,0);
		glRotatef(euler_angles.x, 1,0,0);
		glBegin(GL_QUADS);
			glVertex3f(4, -15, -30);
			glVertex3f(4, 15, -30);
			glVertex3f(4, 15, 0);
			glVertex3f(4, -15, 0);
		glEnd();
	glPopMatrix();
	
	glColor3f(.5f, .5f, .5f);
	glBegin(GL_LINES);
		glVertex3f(1, -15, 0);
		glVertex3f(1, 15, 0);
		glVertex3f(1, 0, -15);
		glVertex3f(1, 0, 15);
	glEnd();

	glPopMatrix();
	glPopAttrib();

	/////////////////////////////////////////////
	// inclination instrument
	glClearColor(0.f, 0.1f, 0.75f, 1.0f );
	glScissor(incl_viewport.left, incl_viewport.bottom, incl_viewport.width, incl_viewport.height);
	glEnable(GL_SCISSOR_TEST);
	glClear(GL_COLOR_BUFFER_BIT);
	glDisable(GL_SCISSOR_TEST);

	incl_viewport.prepare();
	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	glPushMatrix();

	// position the viewer
	inclination.apply();

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glDisable(GL_DEPTH_TEST);

	glColor3f(.0f, .0f, .0f);
		glBegin(GL_LINES);
			glVertex3f(1, -25, 0);
			glVertex3f(1, 25, 0);
			glVertex3f(1, 0, -25);
			glVertex3f(1, 0, 25);
		glEnd();

	glPushMatrix();
		glColor3f(1.0f, 1.0f, 1.f);
		glRotatef(-euler_angles.x, 1,0,0);
		glBegin(GL_LINES);
			glVertex3f(0, -15, 0);
			glVertex3f(0, 15, 0);
			glVertex3f(0, 0, 0);
			glVertex3f(0, 0, 5);
		glEnd();
	glPopMatrix();

	glPopMatrix();
	glPopAttrib();

	/////////////////////////////////////////////
	// instrument pannel
	float x,y,z;
	x = ((float)ctrlState.roll)*4;
	y = ((float)ctrlState.pitch)*4;
	z = ctrlState.thrust*8;

	glClearColor(0.f, 0.1f, 0.75f, 1.0f );
	glScissor(insrt_viewport.left, insrt_viewport.bottom, insrt_viewport.width, insrt_viewport.height);
	glEnable(GL_SCISSOR_TEST);
	glClear(GL_COLOR_BUFFER_BIT);
	glDisable(GL_SCISSOR_TEST);

	insrt_viewport.prepare();
	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	glPushMatrix();

	// position the viewer
	instruments.apply();

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glDisable(GL_DEPTH_TEST);

	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_LINES);
		glVertex3f(10, -5, 0);
		glVertex3f(10, 5, 0);
		glVertex3f(10, 0, -5);
		glVertex3f(10, 0, 5);
	glEnd();
	glColor3f(.25f, .25f, 1.f);
	glBegin(GL_QUADS);
		glVertex3f(10, x+0.5f, y);
		glVertex3f(10, x, y+0.5f);
		glVertex3f(10, x-0.5f, y);
		glVertex3f(10, x, y-0.5f);
	glEnd();

	glColor3f(1.f, 0.f, 0.f);
	glBegin(GL_LINES);
		glVertex3f(10, -0.75f, z-4.f);
		glVertex3f(10, 0.75f, z-4.f);
	glEnd();

	glPopAttrib();
	glPopMatrix();
}
